
-- projects a 3d point onto a 2d surface
function project(point, cam)
  -- convert vect type to column matrix
  local vertex = matrix{{point.x},
                        {point.y},
                        {point.z},
                        {1}}
  
  -- apply view and perspective transforms
  local vertex_view_space = ((cam:perspective_transform() *
                              cam:look_transform()) *
                              cam:location_transform()) *
                              vertex
  
  -- covert column matrix to vect type
  vertex_view_space = vect:new(vertex_view_space[1][1],
                               vertex_view_space[2][1],
                               vertex_view_space[3][1],
                               vertex_view_space[4][1])
  
  -- normalize by dividing through by the homogeneous coordinate w
  local w = vertex_view_space.w
  vertex_view_space.x = vertex_view_space.x / w
  vertex_view_space.y = vertex_view_space.y / w
  vertex_view_space.z = vertex_view_space.z / w
  
  -- if any of .x/.y/.z are outside the range [-1,1], skip this method
  -- commented out because it's more fun when things go offscreen
  --[[
  local v_factor = 4
  if vertex_view_space.x < -v_factor or vertex_view_space.x > v_factor then
    return
  end
  if vertex_view_space.y < -v_factor or vertex_view_space.y > v_factor then
    return
  end
  if vertex_view_space.z < -v_factor or vertex_view_space.z > v_factor then
    return
  end--]]
  
    
  -- now map [-1,1] into the screen coordinates (0,width) and (0,height)
  --  where (0,0) is the top-left corner of the screen
  local x = vertex_view_space.x * (cam.width/2) + (cam.width/2)
  local y = vertex_view_space.y * (cam.height/2) + (cam.height/2)
  
  return vect:new(x, y)
end



--############################################################################--
-- Affine transforms
--############################################################################--

-- translate by an offset tx, ty, tz
function affine_translate(point, tx, ty, tz)

  -- translation transform matrix
  local transform = matrix{{ 1, 0, 0, tx },
                           { 0, 1, 0, ty },
                           { 0, 0, 1, tz },
                           { 0, 0, 0, 1  }}
                           
                           
  -- convert point to a column matrix
  local vertex = matrix{{point.x},
                        {point.y},
                        {point.z},
                        {1      }}
  
  -- apply transformation
  vertex = transform * vertex
  
  -- convert column matric back to a point
  return vect:new(vertex[1][1], vertex[2][1], vertex[3][1])
end


--------------------------------------------------------------------------------
-- translate by scale of tx, ty, tz
function affine_scale(point, sx, sy, sz)
  
  -- scale uniformly if only one scale value given
  if sy == nil or sz == nil then
    sy = sx
    sz = sx
  end

  -- scale transform matrix
  local transform = matrix{{ sx, 0,  0,  0 },
                           { 0,  sy, 0,  0 },
                           { 0,  0,  sz, 0 },
                           { 0,  0,  0,  1 }}
                           
  -- convert point to a column matrix
  local vertex = matrix{{point.x},
                        {point.y},
                        {point.z},
                        {1      }}
  
  -- apply transformation
  vertex = transform * vertex
  
  -- convert column matric back to a point
  return vect:new(vertex[1][1], vertex[2][1], vertex[3][1])
end



--------------------------------------------------------------------------------
-- rotate about the x axis by t
function affine_rotate_x(point, t)
  local sin = math.sin
  local cos = math.cos
  
  -- x rotate transform matrix
  local transform = matrix{{ 1, 0,      0,       0 },
                           { 0, cos(t), -sin(t), 0 },
                           { 0, sin(t), cos(t),  0 },
                           { 0, 0,      0,       1 }}
                           
                           
  -- convert point to a column matrix
  local vertex = matrix{{point.x},
                        {point.y},
                        {point.z},
                        {1      }}
  
  -- apply transformation
  vertex = transform * vertex
  
  -- convert column matric back to a point
  point = vect:new(vertex[1][1], vertex[2][1], vertex[3][1])
  
  return point
end




--------------------------------------------------------------------------------
-- rotate about the y axis by t
function affine_rotate_y(point, t)
  local sin = math.sin
  local cos = math.cos
  
  -- y rotate transform matrix
  local transform = matrix{{ cos(t),  0, sin(t), 0 },
                           { 0,       1, 0,      0 },
                           { -sin(t), 0, cos(t), 0 },
                           { 0,       0, 0,      1 }}
                           
                           
  -- convert point to a column matrix
  local vertex = matrix{{point.x},
                        {point.y},
                        {point.z},
                        {1      }}
  
  -- apply transformation
  vertex = transform * vertex
  
  -- convert column matric back to a point
  point = vect:new(vertex[1][1], vertex[2][1], vertex[3][1])
  
  return point
end




--------------------------------------------------------------------------------
-- rotate about the z axis by t
function affine_rotate_z(point, t)
  local sin = math.sin
  local cos = math.cos
  
  -- z rotate transform matrix
  local transform = matrix{{ cos(t), -sin(t), 0, 0 },
                           { sin(t), cos(t),  0, 0 },
                           { 0,      0,       1, 0 },
                           { 0,      0,       0, 1 }}
                           
                           
  -- convert point to a column matrix
  local vertex = matrix{{point.x},
                        {point.y},
                        {point.z},
                        {1      }}
  
  -- apply transformation
  vertex = transform * vertex
  
  -- convert column matric back to a point
  point = vect:new(vertex[1][1], vertex[2][1], vertex[3][1])
  
  return point
end


--------------------------------------------------------------------------------
-- rotates point about an axis a passing through point c by t degrees
function affine_rotate_axis(point, a, c, t)
  local sin = math.sin
  local cos = math.cos
  
  -- a must be a unit vector
  a = a:unit_vector()

  local transform = matrix{ 
    
      { a.x^2 + (a.y^2 + a.z^2)*cos(t),
        a.x*a.y*(1-cos(t)) - a.z*sin(t),
        a.x*a.z*(1-cos(t)) + a.y*sin(t),
        (c.x*(a.y^2 + a.z^2) - a.x*(c.y*a.y + c.z*a.z))*(1-cos(t)) + (c.y*a.z - c.z*a.y)*sin(t) },
        
      { a.x*a.y*(1-cos(t)) + a.z*sin(t),
        a.y^2 + (a.x^2 + a.z^2)*cos(t),
        a.y*a.z*(1-cos(t)) - a.x*sin(t),
        (c.y*(a.x^2 + a.z^2) - a.y*(c.x*a.x + c.z*a.z))*(1-cos(t)) + (c.z*a.x - c.x*a.z)*sin(t) },
        
      { a.x*a.z*(1-cos(t)) - a.y*sin(t),
        a.y*a.z*(1-cos(t)) + a.x*sin(t),
        a.z^2 + (a.x^2 + a.y^2)*cos(t),
        (c.z*(a.x^2 + a.y^2) - a.z*(c.x*a.x + c.y*a.y))*(1-cos(t)) + (c.x*a.y - c.y*a.x)*sin(t) },
        
      { 0, 0, 0, 1 }
      }
      
  -- convert point to a column matrix
  local vertex = matrix{{point.x},
                        {point.y},
                        {point.z},
                        {1      }}
                        
  -- apply transformation
  vertex = transform * vertex
                         
  -- convert column matrix back to a point
  point = vect:new(vertex[1][1], vertex[2][1], vertex[3][1])    
  
  return point
  
end



  --############################################################################
  -- transform vertices
  --############################################################################

  
  function translate_vertices(vertices, x, y, z)
    local new = {}
    for _,vertex in ipairs(vertices) do
      new[#new+1] = affine_translate(vertex, x, y, z)
    end
    return new
  end

  
  function scale_vertices(vertices, x, y, z)
    local new = {}
    for _,vertex in ipairs(vertices) do
      new[#new+1] = affine_scale(vertex, x, y, z)
    end
    return new
  end
  
  
  function rotate_vertices(vertices, rx, ry, rz)
    local new = {}
    
    --apply x rotation
    if rx ~= 0 and rx ~= nil then
      for _,vertex in ipairs(vertices) do
        new[#new+1] = affine_rotate_x(vertex, rx)
      end
      vertices = new
    end
    
    
    -- apply y rotation
    new = {}
    if ry ~= 0 and ry ~= nil then
      for _,vertex in ipairs(vertices) do
        new[#new+1] = affine_rotate_y(vertex, ry)
      end
      vertices = new
    end
    
    -- apply z rotation
    new = {}
    if rz ~= 0 and rz ~= nil then
      for _,vertex in ipairs(vertices) do
        new[#new+1] = affine_rotate_z(vertex, rz)
      end
      vertices = new
    end
    
    return vertices
  end
  
  
  
  function rotate_axis_vertices(vertices, a, c, t)
    local new = {}
    for _,vertex in ipairs(vertices) do
      new[#new+1] = affine_rotate_axis(vertex, a, c, t)
    end
    return new
  end
  
  
  --############################################################################
  -- useful functions
  --############################################################################
  
  -- given a table of points, return the centroid point
  function get_centroid(points)
    local n = #points
    local sum_x = 0
    local sum_y = 0
    local sum_z = 0
    
    for _,v in ipairs(points) do
      sum_x = sum_x + v.x
      sum_y = sum_y + v.y
      sum_z = sum_z + v.z
    end
    
    sum_x = sum_x/n
    sum_y = sum_y/n
    sum_z = sum_z/n
    
    return vect:new(sum_x, sum_y, sum_z)
  end
  
  
    -- Removes item val from table t
  function remove_item(t, val)
    for i,v in ipairs(t) do
      if v == val then
        table.remove(t, i)
        break
      end
    end
  end
  
  -- returns a table of decimal r, g, b values from a 6 digit hex string
  function hex_to_rgb(hex)
    local color = {}
    local digits = str_to_chars(hex)
    
    color.r = tonumber(digits[1]..digits[2], 16)
    color.g = tonumber(digits[3]..digits[4], 16)
    color.b = tonumber(digits[5]..digits[6], 16)
    
    return color
  end
  
  
  
  --------------------------------------------------------------------------------
  -- returns a table of characters of the string s
  function str_to_chars(s)
    chars = {}
    for i=1,#s do
      chars[i] = string.sub(s, i, i)
    end
    return chars
  end


  --############################################################################
  -- draw functions
  --############################################################################

  -- draws the cooridinate system axis' x, y, z
  function draw_axis()
    axis = {}
    axis.length = 350
    axis.o = vect:new(0, 0, 0)
    axis.x = vect:new(axis.length, 0, 0)
    axis.y = vect:new(0, axis.length, 0)
    axis.z = vect:new(0, 0, axis.length)
  
    -- get projection coordinates
    axiso = project(axis.o, camera)
    axisx = project(axis.x, camera)
    axisy = project(axis.y, camera)
    axisz = project(axis.z, camera)
    
    -- draw
    if axisx ~= nil and axisy ~= nil and axisz ~= nil then
      love.graphics.setColor(0, 0, 200, 255)
      love.graphics.setLine(2, "smooth")
      love.graphics.line(axiso.x, axiso.y, axisx.x, axisx.y)
      love.graphics.line(axiso.x, axiso.y, axisy.x, axisy.y)
      love.graphics.line(axiso.x, axiso.y, axisz.x, axisz.y)
      
      love.graphics.setColor(0, 0, 0, 255)
      love.graphics.setFont(font)
      love.graphics.print("X", axisx.x, axisx.y)
      love.graphics.print("Y", axisy.x, axisy.y)
      love.graphics.print("Z", axisz.x, axisz.y)
    end
    
   
  end

















