

function wireframe_objects_load()

  --############################################################################
  --[[----------------------------------------------------------------------]]--
  -- vect object (a vector)
  --[[----------------------------------------------------------------------]]--
  --############################################################################
  vect = {}
  vect.x = nil
  vect.y = nil
  vect.z = nil
  vect.w = nil
  
  
  vect_mt = { __index = vect }
  function vect:new(x, y, z, w)
    if z == nil then
      z = 0
    end
    if w == nil then
      w = 1
    end
    return setmetatable({x=x, y=y, z=z, w=w}, vect_mt)
  end
  
  
  function vect:length()
    return math.sqrt(self.x^2 + self.y^2 + self.z^2)
  end

  
  -- returns cross product vect of (self x b) 
  function vect:cross_product(b)
    local a = self
    local x = (a.y * b.z) - (a.z * b.y)
    local y = (a.z * b.x) - (a.x * b.z)
    local z = (a.x * b.y) - (a.y * b.x)
    
    return vect:new(x, y, z)
  end
  
  
  -- returns unit vect (vect/|vect|)
  function vect:unit_vector()
    local length = self:length()
    return vect:new(self.x/length, self.y/length, self.z/length)
  end
  
  
  -- returns distance from self to other point
  function vect:distance(other)
    local d = (self.x - other.x)^2 + (self.y - other.y)^2 + (self.z - other.z)^2
    return math.sqrt(d)
  end
  
  
  function vect:copy()
    return vect:new(self.x, self.y, self.z)
  end
  
  
  function vect:print()
    print(self.x.." "..self.y.." "..self.z)
  end
  
  
  
  --############################################################################
  --[[----------------------------------------------------------------------]]--
  -- camera object
  --[[----------------------------------------------------------------------]]--
  --############################################################################
  camera = {}
  camera.position = nil
  camera.u  = nil     -- new x axis
  camera.v  = nil     -- new y axis
  camera.n  = nil     -- new z axis
  
  camera.near = -400
  camera.far = -5000
  camera.width = 800
  camera.height = 600
  
  
  camera_mt = { __index = camera }
  function camera:new(position, look_point)
  
    -- create new z axis (unit vector towards look point)
    local x = look_point.x - position.x
    local y = look_point.y - position.y
    local z = look_point.z - position.z
    local n = vect:new(x, y, z):unit_vector()
    
    -- create new x axis (orthogonal to n and up vector)
    local up = vect:new(0, 1, 0)
    local u = up:cross_product(n):unit_vector()
    
    -- create new y axis (orthogonal n and u vector)
    local v = n:cross_product(u)
  
    return setmetatable({u=u, n=n, v=v, position = position}, camera_mt)
  end
  
  
  -- returns the camera_look_transform matrix
  function camera:look_transform()
    local m = matrix{{ self.u.x, self.u.y, self.u.z, 0 },
                     { self.v.x, self.v.y, self.v.z, 0 },
                     { self.n.x, self.n.y, self.n.z, 0 },
                     { 0,        0,        0,        1 }}
                     
    return m
  end
  
  
  -- rotate camera on it's own x axis (self.u) by t degrees
  function camera:rotate_x(t)
    self.v = affine_rotate_axis(self.v, self.u, vect:new(0,0,0), t)
    self.n = affine_rotate_axis(self.n, self.u, vect:new(0,0,0), t)
    self.position = affine_rotate_axis(self.position, self.u, vect:new(0,0,0), t)
  end
  
  
  -- rotate camera on it's own y axis (self.v) by t degrees
  function camera:rotate_y(t)
    self.u = affine_rotate_axis(self.u, self.v, vect:new(0,0,0), t)
    self.n = affine_rotate_axis(self.n, self.v, vect:new(0,0,0), t)
    self.position = affine_rotate_axis(self.position, self.v, vect:new(0,0,0), t)
  end
  
  
  -- rotate camera on it's own z axis (self.n) by t degrees
  function camera:rotate_z(t)
    self.u = affine_rotate_axis(self.u, self.n, vect:new(0,0,0), t)
    self.v = affine_rotate_axis(self.v, self.n, vect:new(0,0,0), t)
    self.position = affine_rotate_axis(self.position, self.n, vect:new(0,0,0), t)
  end
  
  
  -- translate camera on it's own x axis by tx
  function camera:move_x(tx) 
    local x = tx * self.u.x
    local y = tx * self.u.y
    local z = tx * self.u.z
    
    self.position = affine_translate(self.position, x, y, z)
  end
  
  
  -- translate camera on it's own y axis by tx
  function camera:move_y(ty) 
    local x = ty * self.v.x
    local y = ty * self.v.y
    local z = ty * self.v.z
    
    self.position = affine_translate(self.position, x, y, z)
  end
  
  
  -- translate camera on it's own z axis by tx
  function camera:move_z(tz) 
    local x = tz * self.n.x
    local y = tz * self.n.y
    local z = tz * self.n.z
    
    self.position = affine_translate(self.position, x, y, z)
  end
  
  
  -- returns the camera_location_transform matrix
  function camera:location_transform()
    local m = matrix{{ 1, 0, 0, -self.position.x },
                     { 0, 1, 0, -self.position.y },
                     { 0, 0, 1, -self.position.z },
                     { 0, 0, 0, 1                }}
    return m
  end
  
  
  function camera:perspective_transform()
    local near = self.near
    local far = self.far
    local width = self.width
    local height = self.height
    
    local m = matrix{{ (2*near)/width, 0, 0, 0 },
                     { 0, (2*near)/height, 0, 0 },
                     { 0, 0, -(far+near)/(far-near), -(2*far*near)/(far-near) },
                     { 0, 0, -1, 1                }}
    
    return m
  end

  
  
  --############################################################################
  --[[----------------------------------------------------------------------]]--
  -- Facet object
  --[[----------------------------------------------------------------------]]--
  --############################################################################
  facet = {}
  facet.vertices = nil
  facet.draw_points = nil
  facet.color = nil
  
  facet_mt = { __index = facet }
  function facet:new(points)
    local vertices = {}
    for _,v in ipairs(points) do
      vertices[#vertices+1] = v
    end
    
    local draw_points = {}
  
    local color = {}
    color.line_r = 0
    color.line_g = 0
    color.line_b = 0
    color.line_a = 255
    
    color.body_r = 255
    color.body_g = 0
    color.body_b = 0
    color.body_a = 200
    
    return setmetatable({ vertices = vertices,
                          draw_points = draw_points,
                          color = color }, facet_mt)
  end
  
  ------------------------------------------------------------------------------
  function facet:get_centroid()
    return get_centroid(self.vertices)
  end
  
  
  ------------------------------------------------------------------------------
  function facet:set_draw_points(camera)
    local points = {}
    for _,vertex in ipairs(self.vertices) do
      local newpoint = project(vertex, camera)
      if newpoint ~=nil then
        points[#points+1] = newpoint
      else
        return
      end
    end
    
    self.draw_points = points
  end
  
  
  ------------------------------------------------------------------------------
  function facet:set_color(line, body)
    local c = hex_to_rgb(line)
    self.color.line_r = c.r
    self.color.line_g = c.g
    self.color.line_b = c.b
    
    c = hex_to_rgb(body)
    self.color.body_r = c.r
    self.color.body_g = c.g
    self.color.body_b = c.b
  end
  
  ------------------------------------------------------------------------------
  function facet:draw(line_color, body_color)
    if line_color ~= nil and body_color ~= nil then
      self:set_color(line_color, body_color)
    end
  
    local points = self.draw_points
    
    -- set up line table of points for draw line function
    local line = {}
    for _,v in ipairs(points) do
      line[#line+1] = v.x
      line[#line+1] = v.y
    end
    line[#line+1] = line[1]   -- last point in line is the starting point
    line[#line+1] = line[2]
    
    
    -- fill in the facet
    love.graphics.setColor(self.color.body_r,
                           self.color.body_g,
                           self.color.body_b,
                           self.color.body_a)
    love.graphics.polygon("fill", line)
    
    
    -- draw the vertices
    love.graphics.setColor(self.color.line_r,
                           self.color.line_g,
                           self.color.line_b,
                           self.color.line_a)
    love.graphics.setPoint(4, "smooth")
    for _,point in ipairs(points) do
      love.graphics.point(point.x, point.y)
    end
    
    -- draw the lines
    love.graphics.setLine(2, "smooth")
    love.graphics.line(line)
  end
  
  
  
  
  --############################################################################
  --[[----------------------------------------------------------------------]]--
  -- cube object
  --[[----------------------------------------------------------------------]]--
  --############################################################################
  
  cube = {}
  cube.x = 0            -- location of projected x coordinate
  cube.y = 0            -- location of projected y coordinate
  cube.translate_draw_points = false
  cube.center = nil
  cube.vertices = nil
  cube.facets = nil
  cube.draw_points = nil
  cube.camera = nil
  

  cube_mt = { __index = cube }
  function cube:new(x, y, z, width, height, length, camera)
  
    if x == nil or y == nil or z == nil then
      x, y, z = 0, 0, 0
    end
    if width == nil or height == nil or length == nil then
      width, height, length = 1, 1, 1
    end
    local center = vect:new(x, y, z)
    
    -- set up vertices
    
    -- front
    p1 = vect:new(x - 0.5*width, y + 0.5*height, z + 0.5*length)  -- top left
    p2 = vect:new(x + 0.5*width, y + 0.5*height, z + 0.5*length)  -- top right
    p3 = vect:new(x + 0.5*width, y - 0.5*height, z + 0.5*length)  -- bottom right
    p4 = vect:new(x - 0.5*width, y - 0.5*height, z + 0.5*length)  -- bottom left
    
    -- back
    p5 = vect:new(x - 0.5*width, y + 0.5*height, z - 0.5*length)  -- top left
    p6 = vect:new(x + 0.5*width, y + 0.5*height, z - 0.5*length)  -- top right
    p7 = vect:new(x + 0.5*width, y - 0.5*height, z - 0.5*length)  -- bottom right
    p8 = vect:new(x - 0.5*width, y - 0.5*height, z - 0.5*length)  -- bottom left
    
    local vertices = {p1, p2, p3, p4, p5, p6, p7, p8}
    
    
    -- set up facets
    local facets = {}
    facets[#facets+1] = facet:new({p1, p2, p3, p4})
    facets[#facets+1] = facet:new({p1, p5, p8, p4})
    facets[#facets+1] = facet:new({p5, p6, p7, p8})
    facets[#facets+1] = facet:new({p2, p6, p7, p3})
    facets[#facets+1] = facet:new({p1, p5, p6, p2})
    facets[#facets+1] = facet:new({p8, p7, p3, p4})
    
    -- storage for 2d transformed points
    local draw_points = {}
    draw_points.center = nil
    
    return setmetatable({ center = center,
                          vertices = vertices,
                          facets = facets,
                          draw_points = draw_points,
                          camera = camera}, cube_mt)
  end
  
  
  ------------------------------------------------------------------------------
  function cube:set_center(x, y, z)
    -- subtract center vector from new vector
    diffx = x - self.center.x
    diffy = y - self.center.y
    diffz = z - self.center.z
    
    -- set center
    self.center.x, self.center.y, self.center.z = x, y, z
    
    -- translate vertices
    self.vertices = translate_vertices(self.vertices, diffx, diffy, diffz)
  end
  
  
  ------------------------------------------------------------------------------
  function cube:translate(x, y, z)
    -- translate center
    self.center = affine_translate(self.center, x, y, z)
    
    -- translate vertices
    self.vertices = translate_vertices(self.vertices, x, y, z)
  end
  
  
  ------------------------------------------------------------------------------
  function cube:scale(x, y, z)
    local center_copy = self.center:copy()
    
    
    -- set rectangle at origin
    self:set_center(0, 0, 0)
    
    -- apply scale
    self.vertices = scale_vertices(self.vertices, x, y, z)
    
    -- set rectangle back to original position
    self:set_center(center_copy.x, center_copy.y, center_copy.z )
    
  end


  ------------------------------------------------------------------------------
  function cube:rotate(rx, ry, rz)
  
    local center_copy = self.center:copy()
    
    -- set rectangle at origin
    self:set_center(0, 0, 0)
    
    -- apply rotation transformation
    self.vertices = rotate_vertices(self.vertices, rx, ry, rz)
    
    -- set rectangle back to original position
    self:set_center(center_copy.x, center_copy.y, center_copy.z )    
  end
  
  
  ------------------------------------------------------------------------------
  function cube:rotate_axis(axis, point, r)
    -- apply rotation transformation
    self.vertices = rotate_axis_vertices(self.vertices, axis, point, r)
  end
  
  
  ------------------------------------------------------------------------------
  function cube:update(dt)
    -- setup new facets
    local p1 = self.vertices[1]
    local p2 = self.vertices[2]
    local p3 = self.vertices[3]
    local p4 = self.vertices[4]
    local p5 = self.vertices[5]
    local p6 = self.vertices[6]
    local p7 = self.vertices[7]
    local p8 = self.vertices[8]
    
    local facets = {}
    facets[#facets+1] = facet:new({p1, p2, p3, p4})
    facets[#facets+1] = facet:new({p1, p5, p8, p4})
    facets[#facets+1] = facet:new({p5, p6, p7, p8})
    facets[#facets+1] = facet:new({p2, p6, p7, p3})
    facets[#facets+1] = facet:new({p1, p5, p6, p2})
    facets[#facets+1] = facet:new({p8, p7, p3, p4})
    
    facets[1]:set_color('000000', '00FF00')
    facets[2]:set_color('000000', '0000FF')
    facets[3]:set_color('000000', '00FFFF')
    facets[4]:set_color('000000', 'FFFF00')
    facets[5]:set_color('000000', 'FF00FF')
    
    
    self.facets = facets
    
    -- update draw_points for each facet
    self.draw_points.center = project(self.center, self.camera)
    for _,v in ipairs(self.facets) do
      v:set_draw_points(self.camera)
      
      -- translate drawpoints according to cube.x, cube.y
      if self.translate_draw_points then
       local diffx = self.x - self.draw_points.center.x
       local diffy = self.y - self.draw_points.center.y
      
       v.draw_points = translate_vertices(v.draw_points, diffx, diffy, 0)
     end
    end
    
    
    -- calculate draw order of facets
    
    -- get distance of the centroid of each facet from the camera camera
    for _,v in ipairs(self.facets) do
      local centroid_point = v:get_centroid()
      local camera_point = self.camera.position
      v.distance = camera_point:distance(centroid_point)
    end
    
    -- sort by distance, farthest away from camera first (high to low)
    table.sort(self.facets, function(a,b) return a.distance>b.distance end)
    
  end
  
  
  
  ------------------------------------------------------------------------------
  function cube:draw()
    -- draw facets
    for i,v in ipairs(self.facets) do
      v:draw()
    end
  end
  
  
end


































